from adagio.shells.interfaceless import function_to_taskspec, _get_origin_type, _parse_annotation
from typing import Any, Dict, List, Union, Optional, Tuple
from pytest import raises
import inspect


def test_function_to_taskspec():
    ts = function_to_taskspec(f1, lambda ds: [d["data_type"] is not int for d in ds])
    assert 2 == len(ts.inputs)
    assert "a" in ts.inputs and "b" in ts.inputs
    assert all(r.required for r in ts.inputs.values())
    assert ts.inputs["a"].nullable
    assert not ts.inputs["b"].nullable

    assert 2 == len(ts.configs)
    assert "c" in ts.configs and "d" in ts.configs
    assert not ts.configs["c"].nullable
    assert ts.configs["c"].required
    assert ts.configs["d"].nullable
    assert not ts.configs["d"].required
    assert ts.configs["d"].default_value == "x"
    assert all(r.data_type is str for r in ts.configs.values())

    assert 1 == len(ts.outputs)
    assert "_0" in ts.outputs
    assert not ts.outputs["_0"].nullable
    assert ts.outputs["_0"].data_type is int

    ts = function_to_taskspec(f2, lambda ds: [d["data_type"] is not int for d in ds],
                              deterministic=False, lazy=True)
    assert ts.lazy
    assert not ts.deterministic

    assert 2 == len(ts.inputs)
    assert "a" in ts.inputs and "c" in ts.inputs
    assert 1 == len(ts.configs)
    assert "b" in ts.configs
    assert ts.configs["b"].data_type is object
    assert ts.configs["b"].nullable
    assert 2 == len(ts.outputs)
    assert "_0" in ts.outputs and "_1" in ts.outputs
    assert not ts.outputs["_0"].nullable
    assert ts.outputs["_1"].nullable

    ts = function_to_taskspec(f3, lambda ds: [d["data_type"] is not int for d in ds])
    print(ts.to_json(True))
    assert 0 == len(ts.outputs)

    ts = function_to_taskspec(f4, lambda ds: [d["data_type"] is not int for d in ds])
    print(ts.to_json(True))
    assert 0 == len(ts.outputs)

    ts = function_to_taskspec(f5, lambda ds: [d["data_type"] is not int for d in ds])
    print(ts.to_json(True))
    assert 0 == len(ts.inputs)
    assert 0 == len(ts.outputs)

    # TODO: not tested on func and metadata because that is supposed to change to something else


def test__parse_annotation():
    assert dict(data_type=object, nullable=True) == _parse_annotation(None)
    assert dict(data_type=object, nullable=True) == _parse_annotation(
        inspect.Parameter.empty)
    assert dict(data_type=object, nullable=True) == _parse_annotation(Any)
    assert dict(data_type=int, nullable=False) == _parse_annotation(int)
    assert dict(data_type=dict, nullable=False) == _parse_annotation(Dict[str, Any])
    assert dict(data_type=object, nullable=True) == _parse_annotation(
        Optional[Any])
    assert dict(data_type=str, nullable=True) == _parse_annotation(
        Optional[str])
    assert dict(data_type=dict, nullable=True) == _parse_annotation(
        Optional[Dict[str, Any]])
    assert dict(data_type=dict, nullable=True) == _parse_annotation(
        Union[None, Dict[str, Any]])
    assert dict(data_type=dict, nullable=True) == _parse_annotation(
        Union[Dict[str, Any], None])
    assert dict(data_type=dict, nullable=True) == _parse_annotation(
        Union[Dict[str, Any], None, None])
    assert dict(data_type=dict, nullable=False) == _parse_annotation(
        Union[Dict[str, Any]])
    raises(TypeError, lambda: _parse_annotation(Union[Dict[str, Any], List[str]]))
    raises(TypeError, lambda: _parse_annotation(Union[Dict[str, Any], List[str], None]))
    raises(TypeError, lambda: _parse_annotation(Union[None]))
    raises(TypeError, lambda: _parse_annotation(Union[None, None]))
    raises(TypeError, lambda: _parse_annotation(type(None)))


def test__get_origin_type():
    assert _get_origin_type(Any) is object
    assert _get_origin_type(Dict[str, Any]) is dict
    assert _get_origin_type(List[str]) is list
    assert _get_origin_type(List[Any]) is list
    assert _get_origin_type(Tuple[int, str]) is tuple
    assert _get_origin_type(Union[int, str], False) is Union
    assert _get_origin_type(Union[None]) is type(None)
    assert _get_origin_type(int) is int


def f1(a: Optional[int], b: "int", c: str, d: "Optional[str]" = "x") -> "int":
    return a + b


def f2(a: int, b, c: int) -> "Tuple[int,Optional[str]]":
    return a + c


def f3(a: int, b: str, c: int):
    pass


def f4(a: int, b: str, c: int) -> None:
    pass


def f5():
    pass



import inspect
import typing
from typing import Any, Callable, Dict, List, Optional, Union, get_type_hints

from adagio.instances import TaskContext
from adagio.specs import ConfigSpec, InputSpec, OutputSpec, TaskSpec
from triad.utils.assertion import assert_or_throw
from triad.utils.convert import to_function, get_full_type_path


def function_to_taskspec(
    func: Callable,
    is_config: Callable[[List[Dict[str, Any]]], List[bool]],
    deterministic: bool = True,
    lazy: bool = False,
) -> TaskSpec:
    specs = inspect.getfullargspec(func)
    sig = inspect.signature(func)
    annotations = get_type_hints(func)
    assert_or_throw(
        specs.varargs is None and specs.varkw is None and len(specs.kwonlyargs) == 0,
        "Function can't have varargs or kwargs",
    )
    inputs: List[InputSpec] = []
    configs: List[ConfigSpec] = []
    outputs: List[OutputSpec] = []
    arr: List[Dict[str, Any]] = []
    for k, w in sig.parameters.items():
        anno = annotations.get(k, w.annotation)
        a = _parse_annotation(anno)
        a["name"] = k
        if w.default == inspect.Parameter.empty:
            a["required"] = True
        else:
            a["required"] = False
            a["default_value"] = w.default
        arr.append(a)
    cfg = is_config(arr)
    for i in range(len(cfg)):
        if cfg[i]:
            configs.append(ConfigSpec(**arr[i]))
        else:
            assert_or_throw(
                arr[i]["required"], f"{arr[i]}: dependency must not have default value"
            )
            inputs.append(InputSpec(**arr[i]))
    n = 0
    anno = annotations.get("return", sig.return_annotation)
    is_multiple = _is_tuple(anno)
    for x in list(anno.__args__) if is_multiple else [anno]:
        if x == inspect.Parameter.empty or x is type(None):  # noqa: E721
            continue
        a = _parse_annotation(x)
        a["name"] = f"_{n}"
        outputs.append(OutputSpec(**a))
        n += 1
    metadata = dict(__interfaceless_func=get_full_type_path(func))
    return TaskSpec(
        configs,
        inputs,
        outputs,
        _interfaceless_wrapper,
        metadata,
        deterministic=deterministic,
        lazy=lazy,
    )


def _interfaceless_wrapper(ctx: TaskContext) -> None:
    ctx.ensure_all_ready()
    func = to_function(ctx.metadata.get_or_throw("__interfaceless_func", object))
    o = func(**ctx.inputs, **ctx.configs)
    res = list(o) if isinstance(o, tuple) else [o]
    n = 0
    for k in ctx.outputs.keys():
        ctx.outputs[k] = res[n]
        n += 1


def _parse_annotation(anno: Any) -> Dict[str, Any]:
    d = _try_parse(anno)
    if d is not None:
        return d
    assert_or_throw(
        anno.__module__ == "typing", TypeError(f"{anno} is not a valid type")
    )
    nullable = False
    if _is_union(anno):
        tps = set(anno.__args__)
        if type(None) not in tps or len(tps) != 2:
            raise TypeError(f"{anno} can't be converted for TaskSpec")
        anno = [x for x in tps if x is not type(None)][0]  # noqa: E721
        nullable = True
    d = _try_parse(anno)
    if d is not None:
        d["nullable"] = nullable
        return d
    return dict(data_type=_get_origin_type(anno), nullable=nullable)


def _try_parse(anno: Any) -> Optional[Dict[str, Any]]:
    if anno is None or anno == inspect.Parameter.empty or anno is Any:
        return dict(data_type=object, nullable=True)
    if _is_native_type(anno):
        assert_or_throw(
            anno is not type(None),  # noqa: E721
            TypeError(f"{anno} NoneType is invalid"),
        )
        return dict(data_type=anno, nullable=False)
    return None


def _get_origin_type(anno: Any, assert_is_type: bool = True) -> Any:
    # this part looks weird but it's working for 3.6-3.8
    if anno is not None and anno.__module__ == "typing":
        if anno is Any:
            return object
        if hasattr(typing, "get_origin"):  # pragma: no cover
            anno = typing.get_origin(anno)  # type: ignore  # 3.8
        elif hasattr(anno, "__extra__"):  # pragma: no cover
            anno = anno.__extra__  # < 3.7
        elif hasattr(anno, "__origin__"):  # pragma: no cover
            anno = anno.__origin__  # 3.7
    if anno is None:  # pragma: no cover
        anno = type(None)
    if assert_is_type:
        assert_or_throw(
            _is_native_type(anno), TypeError(f"Can't find python type for {anno}")
        )
    return anno


def _is_native_type(anno: Any) -> bool:
    return isinstance(anno, type) and anno.__module__ != "typing"


def _is_tuple(anno: Any):
    return _get_origin_type(anno) is tuple


def _is_union(anno: Any):
    return _get_origin_type(anno, False) is Union



import threading
import time
from collections import OrderedDict
from threading import RLock
from time import sleep
from typing import Any, Tuple

from adagio.exceptions import AbortedError
from adagio.instances import (NoOpCache, ParallelExecutionEngine, TaskContext,
                              WorkflowContext, WorkflowHooks,
                              WorkflowResultCache, _ConfigVar, _Input,
                              _make_top_level_workflow, _Output, _State, _Task,
                              _Workflow)
from adagio.shells.interfaceless import function_to_taskspec
from adagio.specs import InputSpec, OutputSpec, WorkflowSpec, _NodeSpec
from pytest import raises
from triad.exceptions import InvalidOperationError
from timeit import timeit


def test_task_skip():
    ts = build_task(example_helper1, t1, inputs=dict(a=1), configs=dict(b="xx"))
    assert _State.CREATED == ts.state
    ts.skip()
    assert _State.SKIPPED == ts.state
    raises(InvalidOperationError, lambda: ts.skip())
    ts.run()  # run is no op after skip
    assert _State.SKIPPED == ts.state


def test_task_run_finished():
    ts = build_task(example_helper1, t2, inputs=dict(a=1), configs=dict(b="xx"))
    assert _State.CREATED == ts.state
    ts.run()
    assert _State.FINISHED == ts.state
    assert ts.outputs["_0"].is_successful
    assert 3 == ts.outputs["_0"].value
    assert ts.outputs["_1"].is_successful
    assert -1 == ts.outputs["_1"].value


def test_task_run_finished_native():
    ts = build_task(example_helper1, None, inputs=dict(a=1), configs=dict(b="xx"))
    assert _State.CREATED == ts.state
    ts.run()
    assert _State.FINISHED == ts.state
    assert ts.outputs["_0"].is_successful
    assert 3 == ts.outputs["_0"].value
    assert ts.outputs["_1"].is_successful
    assert -1 == ts.outputs["_1"].value


def test_task_run_cached():
    cache = MockCache()
    ts = build_task(example_helper1, t2, inputs=dict(a=1),
                    configs=dict(b="xx"), cache=cache)
    id1 = ts.__uuid__()
    oid1 = ts.outputs["_0"].__uuid__()
    assert 1 == cache.get_called  # tried first output, not found, quit
    assert 0 == cache.set_called
    assert 0 == cache.skip_called
    assert _State.CREATED == ts.state
    ts.run()
    assert _State.FINISHED == ts.state
    assert ts.outputs["_0"].is_successful
    assert 3 == ts.outputs["_0"].value
    assert ts.outputs["_1"].is_successful
    assert -1 == ts.outputs["_1"].value
    assert 1 == cache.get_called
    assert 2 == cache.set_called  # set both outputs
    assert 0 == cache.skip_called
    assert id1 == ts.__uuid__()
    assert oid1 == ts.outputs["_0"].__uuid__()

    assert 2 == len(cache.tb)
    ts = build_task(example_helper1, t2, inputs=dict(a=1),
                    configs=dict(b="xx"), cache=cache)
    assert id1 == ts.__uuid__()
    assert oid1 == ts.outputs["_0"].__uuid__()
    assert 3 == cache.get_called  # got both cached results
    assert 2 == cache.set_called
    assert 0 == cache.skip_called
    assert _State.FINISHED == ts.state
    assert ts.outputs["_0"].is_successful
    assert 3 == ts.outputs["_0"].value
    assert ts.outputs["_1"].is_successful
    assert -1 == ts.outputs["_1"].value


def test_task_run_cached_with_skip():
    cache = MockCache()
    ts = build_task(example_helper1, t1, inputs=dict(a=1),
                    configs=dict(b="xx"), cache=cache)
    id1 = ts.__uuid__()
    oid1 = ts.outputs["_0"].__uuid__()
    assert 1 == cache.get_called  # tried first output, not found, quit
    assert 0 == cache.set_called
    assert 0 == cache.skip_called
    assert _State.CREATED == ts.state
    ts.run()
    assert _State.FINISHED == ts.state
    assert ts.outputs["_0"].is_successful
    assert 3 == ts.outputs["_0"].value
    assert ts.outputs["_1"].is_skipped
    assert 1 == cache.get_called
    assert 1 == cache.set_called
    assert 1 == cache.skip_called
    assert id1 == ts.__uuid__()
    assert oid1 == ts.outputs["_0"].__uuid__()

    assert 2 == len(cache.tb)
    ts = build_task(example_helper1, t1, inputs=dict(a=1),
                    configs=dict(b="xx"), cache=cache)
    assert id1 == ts.__uuid__()
    assert oid1 == ts.outputs["_0"].__uuid__()
    assert 3 == cache.get_called  # got both cached results
    assert 1 == cache.set_called
    assert 1 == cache.skip_called
    assert _State.FINISHED == ts.state
    assert ts.outputs["_0"].is_successful
    assert 3 == ts.outputs["_0"].value
    assert ts.outputs["_1"].is_skipped


def test_task_run_cached_non_deterministic():
    cache = MockCache()
    ts = build_task(example_helper1, t2, inputs=dict(a=1),
                    configs=dict(b="xx"), cache=cache, deterministic=False)
    id1 = ts.__uuid__()
    oid1 = ts.outputs["_0"].__uuid__()
    assert 0 == cache.get_called
    assert 0 == cache.set_called
    assert 0 == cache.skip_called
    assert _State.CREATED == ts.state
    ts.run()
    assert _State.FINISHED == ts.state
    assert ts.outputs["_0"].is_successful
    assert 3 == ts.outputs["_0"].value
    assert ts.outputs["_1"].is_successful
    assert -1 == ts.outputs["_1"].value
    assert 0 == cache.get_called
    assert 0 == cache.set_called  # set both outputs
    assert 0 == cache.skip_called
    assert id1 == ts.__uuid__()
    assert oid1 == ts.outputs["_0"].__uuid__()

    assert 0 == len(cache.tb)
    ts = build_task(example_helper1, t2, inputs=dict(a=1),
                    configs=dict(b="xx"), cache=cache, deterministic=False)
    assert id1 != ts.__uuid__()
    assert oid1 != ts.outputs["_0"].__uuid__()
    assert 0 == cache.get_called  # got both cached results
    assert 0 == cache.set_called
    assert 0 == cache.skip_called
    assert _State.CREATED == ts.state


def test_task_run_failed():
    ts = build_task(example_helper1, t3, inputs=dict(a=1), configs=dict(b="xx"))
    assert _State.CREATED == ts.state
    ts.run()
    assert _State.FAILED == ts.state
    assert isinstance(ts._exception, SyntaxError)
    assert ts.outputs["_0"].is_successful
    assert 3 == ts.outputs["_0"].value
    assert ts.outputs["_1"].is_failed
    assert isinstance(ts.outputs["_1"].exception, SyntaxError)


def test_task_run_aborted():
    # Abort before run will skip
    ts = build_task(example_helper1, t3, inputs=dict(a=1), configs=dict(b="xx"))
    assert _State.CREATED == ts.state
    ts.ctx.abort()
    ts.run()
    assert _State.SKIPPED == ts.state
    assert ts.outputs["_0"].is_skipped
    assert ts.outputs["_1"].is_skipped
    # Abort during run will abort
    ts = build_task(example_helper1, t4, inputs=dict(a=1), configs=dict(b="xx"))
    assert _State.CREATED == ts.state
    x = threading.Thread(target=ts.run)
    x.start()
    time.sleep(0.1)
    ts.ctx.abort()
    x.join()
    assert _State.ABORTED == ts.state
    assert ts.outputs["_0"].is_successful
    assert 3 == ts.outputs["_0"].value
    assert ts.outputs["_1"].is_failed
    assert isinstance(ts.outputs["_1"].exception, AbortedError)


def test_workflow_build():
    """
       a   d       j
       |   |      / \
       b   e     |   k
       |   |     |   |
       c   |     aa  bb <----- sub start
        \ /      |   |  \
         f       a   b   |
         |       |  /|   |
         g       | | c   |
        / \      |  \    |
       h   i     cc  dd  ee <----- sub end
    """
    s1 = SimpleSpec(["aa", "bb"], ["cc", "dd", "ee"])
    s1.add("a", example_helper_task1, "*aa")
    s1.add("b", example_helper_task1, "*bb")
    s1.add("c", example_helper_task1)
    s1.link("cc", "a._0")
    s1.link("dd", "b._0")
    s1.link("ee", "bb")
    s = SimpleSpec()
    s.add("a", example_helper_task0)
    s.add("b", example_helper_task1)
    s.add("c", example_helper_task1)
    s.add("d", example_helper_task0)
    s.add("e", example_helper_task1)
    s.add("f", example_helper_task2, "c", "e")
    s.add("g", example_helper_task3)
    s.add("h", example_helper_task3, "g._0")
    s.add("i", example_helper_task3, "g._1")
    s.add("j", example_helper_task0)
    s.add("k", example_helper_task1)
    s.add_task("x", s1, {"aa": "j._0", "bb": "k._0"})
    ctx = WorkflowContext()
    raises(InvalidOperationError, lambda: _make_top_level_workflow(s1, ctx, {}))
    wf = _make_top_level_workflow(s, ctx, {})
    assert wf.tasks["a"].__uuid__() == wf.tasks["d"].__uuid__()
    assert wf.tasks["a"].execution_id != wf.tasks["d"].execution_id
    assert {wf.tasks["j"]} == wf.tasks["x"].tasks["a"].upstream

    #assert 14 == len(ctx._tasks)

    def assert_dep(node, up, down):
        assert set(list(up)) == (up if isinstance(up, set) else set(
            t.name for t in wf.tasks[node].upstream))
        assert set(list(down)) == (down if isinstance(down, set) else set(
            t.name for t in wf.tasks[node].downstream))

    assert_dep("a", "", "b")
    assert_dep("b", "a", "c")
    assert_dep("c", "b", "f")
    assert_dep("d", "", "e")
    assert_dep("e", "d", "f")
    assert_dep("f", "ce", "g")
    assert_dep("g", "f", "hi")
    assert_dep("h", "g", "")
    assert_dep("i", "g", "")
    assert_dep("j", "", "ak")  # downstream are always tasks no workflow (x)
    assert_dep("k", "j", "b")
    assert_dep("x", "jk", "")


def test_workflow_run():
    """
       a   d       j
       |   |      / \
       b   e     |   k
       |   |     |   |
       c   |     aa  bb <----- sub start
        \ /      |   |  \
         f       _a  _b  |
         |       |  /|   |
         g       | | _c  |
        / \      |  \    |
       h   i     cc  dd  ee <----- sub end
                     |   |
                     l   m
    """
    s1 = SimpleSpec(["aa", "bb"], ["cc", "dd", "ee"])
    s1.add("_a", example_helper_task1, "*aa")
    s1.add("_b", example_helper_task1, "*bb")
    s1.add("_c", example_helper_task1)
    s1.link("cc", "_a._0")
    s1.link("dd", "_b._0")
    s1.link("ee", "bb")
    s = SimpleSpec()
    s.add("a", example_helper_task0)
    s.add("b", example_helper_task1)
    s.add("c", example_helper_task1)
    s.add("d", example_helper_task0)
    s.add("e", example_helper_task1)
    s.add("f", example_helper_task2, "c", "e")
    s.add("g", example_helper_task3)
    s.add("h", example_helper_task1, "g._0")
    s.add("i", example_helper_task1, "g._1")
    s.add("j", example_helper_task0)
    s.add("k", example_helper_task1)
    s.add_task("x", s1, {"aa": "j._0", "bb": "k._0"})
    s.add("l", example_helper_task1, "x.dd")
    s.add("m", example_helper_task1, "x.ee")
    hooks = MockHooks(None)
    ctx = WorkflowContext(hooks=hooks)
    ctx.run(s, {})
    expected = {'a': 10, 'b': 11, 'c': 12, 'd': 10, 'e': 11, 'f': 23,
                'h': 34, 'i': 14, 'j': 10, 'k': 11, '_a': 11, '_b': 12,
                '_c': 13, 'l': 13, 'm': 12}
    for k, v in expected.items():
        assert v == hooks.res[k]


def test_workflow_run_parallel():
    # simplest case
    s = SimpleSpec()
    s.add("a", wait_task0)
    s.add("b", wait_task0)
    hooks = MockHooks(None)
    ctx = WorkflowContext(hooks=hooks)
    ctx._engine = ParallelExecutionEngine(2, ctx)
    ctx.run(s, {})
    expected = {'a': 1, 'b': 1}
    for k, v in expected.items():
        assert v == hooks.res[k]
    # with exception
    s = SimpleSpec()
    s.add("a", wait_task0)
    s.add("b", wait_task0e)
    hooks = MockHooks(None)
    ctx = WorkflowContext(hooks=hooks)
    ctx._engine = ParallelExecutionEngine(2, ctx)
    with raises(NotImplementedError):
        ctx.run(s, {})
    expected = {'a': 1}
    for k, v in expected.items():
        assert v == hooks.res[k]
    # on failure should cause abort
    s = SimpleSpec()
    s.add("a", wait_task0)
    s.add("b", wait_task0e)
    s.add("c", wait_task1, "a")
    s.add("d", wait_task1, "b")
    s.add("e", wait_task1, "d")
    s.add("f", wait_task1, "c")
    hooks = MockHooks(None)
    ctx = WorkflowContext(hooks=hooks)
    ctx._engine = ParallelExecutionEngine(10, ctx)

    def run():
        with raises(NotImplementedError):
            ctx.run(s, {})

    t = timeit(run, number=1)
    assert t < 0.2   # only a and b are executed

    expected = {'a': 1}
    for k, v in expected.items():
        assert v == hooks.res[k]
    assert "b" in hooks.failed

    # order of execution
    s = SimpleSpec()
    s.add("a", wait_task0)
    s.add("b", wait_task0)
    s.add("c", wait_task1, "a")
    s.add("d", wait_task1, "b")
    s.add("e", wait_task1, "d")
    s.add("f", wait_task1, "c")
    hooks = MockHooks(None)
    ctx = WorkflowContext(hooks=hooks)
    ctx._engine = ParallelExecutionEngine(2, ctx)
    t = timeit(lambda: ctx.run(s, {}), number=1)
    assert t < 0.4
    assert 3 == hooks.res["e"]
    assert 3 == hooks.res["f"]


def test_workflow_run_with_exception():
    s = SimpleSpec()
    s.add("a", example_helper_task0)
    s.add("b", example_helper_task1e)
    s.add("c", example_helper_task1)
    ctx = WorkflowContext()
    raises(NotImplementedError, lambda: ctx.run(s, {}))


def test_workflow_run_with_cache():
    s = SimpleSpec()
    s.add("a", example_helper_task0)
    s.add("c", example_helper_task1)
    cache = MockCache()
    hooks1 = MockHooks(None)
    ctx = WorkflowContext(cache=cache, hooks=hooks1)
    ctx.run(s, {})
    assert 2 == cache.get_called
    assert 0 == cache.hit
    assert 2 == cache.set_called
    expected = {"a": 10, "c": 11}
    for k, v in expected.items():
        assert v == hooks1.res[k]
    # for the second run, get cache will be called, set will not
    ctx.run(s, {})
    assert 4 == cache.get_called
    assert 2 == cache.hit
    assert 2 == cache.set_called
    expected = {"a": 10, "c": 11}
    for k, v in expected.items():
        assert v == hooks1.res[k]
    # for the third run, get cache will be called, set will not
    hooks2 = MockHooks(None)
    ctx = WorkflowContext(cache=cache, hooks=hooks2)
    ctx.run(s, {})
    assert 6 == cache.get_called
    assert 4 == cache.hit
    assert 2 == cache.set_called
    expected = {"a": 10, "c": 11}
    for k, v in expected.items():
        assert v == hooks1.res[k]
    # TODO: skip is not tested


def t1(ctx: TaskContext):
    a = ctx.inputs.get_or_throw("a", int)
    b = ctx.configs.get_or_throw("b", str)
    ctx.outputs["_0"] = a + len(b)


def t2(ctx: TaskContext):
    ctx.ensure_all_ready()
    a = ctx.inputs.get_or_throw("a", int)
    b = ctx.configs.get_or_throw("b", str)
    ctx.outputs["_0"] = a + len(b)
    ctx.outputs["_1"] = a - len(b)
    ctx.log.info("done")


def t3(ctx: TaskContext):
    a = ctx.inputs.get_or_throw("a", int)
    b = ctx.configs.get_or_throw("b", str)
    ctx.outputs["_0"] = a + len(b)
    raise SyntaxError("Expected")


def t4(ctx: TaskContext):
    a = ctx.inputs.get_or_throw("a", int)
    b = ctx.configs.get_or_throw("b", str)
    ctx.outputs["_0"] = a + len(b)
    for i in range(5):
        time.sleep(0.1)
        if ctx.abort_requested:
            raise AbortedError()
    raise SyntaxError("Expected")


def build_task(example_func, func=None, inputs=None,
               configs=None, cache="NoOpCache",
               deterministic=True, task_name="taskname"):
    ts = function_to_taskspec(example_func, lambda ds: [
        d["data_type"] is str for d in ds])
    ns = _NodeSpec(None, task_name, {}, {}, {})
    ts._node_spec = ns
    if func is not None:
        ts.func = func
    ts.deterministic = deterministic
    wfctx = WorkflowContext(cache=cache)
    t = _Task(ts, wfctx)
    if inputs is not None:
        for k, v in inputs.items():
            t.inputs[k].dependency = 1  # set a dummy value so will not complaint
            t.inputs[k]._cached = True
            t.inputs[k]._cached_value = v
    if configs is not None:
        for k, v in configs.items():
            t.configs[k].set(v)
    t.update_by_cache()
    return t


class SimpleSpec(WorkflowSpec):
    def __init__(self, inputs=[], outputs=[]):
        ip = [InputSpec(x, int, False) for x in inputs]
        op = [OutputSpec(x, int, False) for x in outputs]
        super().__init__(inputs=ip, outputs=op)
        self.cursor = None

    def add(self, name, func, *dep):
        ts = function_to_taskspec(func, lambda ds: [
            d["data_type"] is str for d in ds])
        dependency = {}
        if len(ts.inputs) > 0:
            if len(ts.inputs) == 1:
                if len(dep) == 0:
                    dep = [self.cursor.name]
            for f, t in zip(ts.inputs.keys(), dep):
                if t.startswith("*"):
                    t = t[1:]
                elif "." not in t:
                    t = t + "." + self.tasks[t].outputs.get_key_by_index(0)
                dependency[f] = t
        self.cursor = self.add_task(name, ts, dependency=dependency)


class MockCache(WorkflowResultCache):
    def __init__(self, ctx=None):
        self.tb = dict()
        self.set_called = 0
        self.skip_called = 0
        self.get_called = 0
        self.hit = 0

    def set(self, key: str, value: Any) -> None:
        self.tb[key] = (False, value)
        print("set", key)
        self.set_called += 1

    def skip(self, key: str) -> None:
        self.tb[key] = (True, None)
        self.skip_called += 1

    def get(self, key: str) -> Tuple[bool, bool, Any]:
        self.get_called += 1
        if key not in self.tb:
            print("not get", key)
            return False, False, None
        x = self.tb[key]
        print("get", key)
        self.hit += 1
        return True, x[0], x[1]


def example_helper1(a: int, b: str = "x") -> Tuple[int, int]:
    return 1 + len(b), 1 - len(b)


def example_helper_task0() -> int:
    return 10


def example_helper_task1(a: int) -> int:
    return a + 1


def example_helper_task1e(a: int) -> int:
    raise NotImplementedError


def example_helper_task2(a: int, b: int) -> int:
    return a + b


def example_helper_task3(a: int) -> Tuple[int, int]:
    return a + 10, a - 10


def wait_task0() -> int:
    sleep(0.1)
    return 1


def wait_task0e() -> int:
    raise NotImplementedError


def wait_task1(a: int) -> int:
    sleep(0.1)
    return a + 1


def wait_task2(a: int) -> None:
    sleep(0.1)
    print(a + 1)


class MockHooks(WorkflowHooks):
    def __init__(self, wf_ctx: "WorkflowContext"):
        super().__init__(wf_ctx)
        self.lock = RLock()
        self.res = OrderedDict()
        self.skipped = set()
        self.failed = set()

    def on_task_change(
        self,
        task: "_Task",
        old_state: "_State",
        new_state: "_State",
        e=None
    ):
        if new_state == _State.FINISHED and len(task.outputs) == 1:
            with self.lock:
                self.res[task.name] = task.outputs["_0"].value
        if new_state == _State.SKIPPED:
            self.skipped.add(task.name)
        if new_state == _State.FAILED:
            self.failed.add(task.name)



import json
from typing import Callable, cast

from adagio.exceptions import SkippedError
from adagio.instances import (_ConfigVar, _Dependency, _DependencyDict, _Input,
                              _Output, _Task)
from adagio.specs import ConfigSpec, InputSpec, OutputSpec, TaskSpec
from pytest import raises
from triad.collections.dict import IndexedOrderedDict, ParamDict
from triad.exceptions import InvalidOperationError
from triad.utils.hash import to_uuid


def test_dependency():
    a = _Dependency()
    b = _Dependency().set_dependency(a)
    c = _Dependency().set_dependency(b)
    d = _Dependency().set_dependency(c)
    assert d.dependency is a


def test_output():
    t = MockTaskForVar()
    s = OutputSpec("o", dict, False)
    o = _Output(t, s)
    assert to_uuid(t, s) == o.__uuid__()
    assert not o.is_set
    assert not o.is_skipped
    assert not o.is_successful
    assert not o.is_failed
    raises(ValueError, lambda: o.set(1))
    assert o.is_set
    assert not o.is_skipped
    assert not o.is_successful
    assert o.is_failed
    assert isinstance(o.exception, ValueError)
    assert o.trace is not None
    o.set(dict())  # when is_set, setting again will do nothing
    assert o.is_set
    assert not o.is_skipped
    assert not o.is_successful
    assert o.is_failed
    assert isinstance(o.exception, ValueError)

    o = _Output(t, s)
    # setting a bad value will cause exception on both setters and getters
    raises(ValueError, lambda: o.set(None))
    assert o.is_set
    assert not o.is_skipped
    assert not o.is_successful
    assert o.is_failed
    assert isinstance(o.exception, ValueError)

    s2 = OutputSpec("o", dict, True)
    o = _Output(t, s2)
    o.set(None)
    assert o.is_set
    assert not o.is_skipped
    assert o.is_successful
    assert not o.is_failed
    assert o.exception is None

    s2 = OutputSpec("o", dict, True)
    o = _Output(t, s2)
    o.skip()
    assert o.is_set
    assert o.is_skipped
    assert not o.is_successful
    assert not o.is_failed
    assert o.exception is None


def test_input():
    t = MockTaskForVar()
    s = OutputSpec("o", dict, False)
    o = _Output(t, s)
    p = ParamDict()
    ii = InputSpec("x", dict, False, False, default_value=p, default_on_timeout=True)
    i = _Input(t, ii)
    i.set_dependency(o)
    raises(ValueError, lambda: o.set(None))
    raises(ValueError, lambda: i.get())

    t = MockTaskForVar()
    s = OutputSpec("o", ParamDict, False)
    o = _Output(t, s)
    raises(AssertionError, lambda: InputSpec("x", dict, False, False,
                                             timeout="0.1s",
                                             default_value=None,
                                             default_on_timeout=True))

    # Input linked with Output
    t = MockTaskForVar()
    s = OutputSpec("o", ParamDict, False)
    o = _Output(t, s)
    p = ParamDict()
    p2 = ParamDict()
    ii = InputSpec("x", dict, False, False, timeout="0.1s",
                   default_value=p, default_on_timeout=True)
    i = _Input(t, ii).set_dependency(o)
    assert p is i.get()
    o.set(p2)
    assert p is not i.get()
    assert p2 is i.get()
    # Input linked with Input
    i2 = _Input(t, ii).set_dependency(i)
    assert p is not i2.get()
    assert p2 is i2.get()

    t = MockTaskForVar()
    s = OutputSpec("o", ParamDict, False)
    o = _Output(t, s)
    p = ParamDict()
    p2 = ParamDict()
    ii = InputSpec("x", dict, False, False, timeout="0.1s",
                   default_value=p, default_on_timeout=False)
    i = _Input(t, ii).set_dependency(o)
    raises(TimeoutError, lambda: i.get())

    # Output skipped, input without default will raise error
    t = MockTaskForVar()
    s = OutputSpec("o", ParamDict, False)
    o = _Output(t, s)
    p = ParamDict()
    ii = InputSpec("x", dict, False)
    i = _Input(t, ii).set_dependency(o)
    o.skip()
    raises(SkippedError, lambda: i.get())

    # Output skipped, input with default will return default
    t = MockTaskForVar()
    s = OutputSpec("o", ParamDict, False)
    o = _Output(t, s)
    p = ParamDict()
    ii = InputSpec("x", dict, False, False, p)
    i = _Input(t, ii).set_dependency(o)
    o.skip()
    assert p is i.get()

    # Output -> workflow output -> Input
    t = MockTaskForVar()
    s = OutputSpec("o", ParamDict, False)
    oo = _Output(t, s)  # task output
    o = _Output(t, s)  # workflow output
    o.set_dependency(oo)
    p = ParamDict()
    ii = InputSpec("x", dict, False)
    i = _Input(t, ii).set_dependency(o)
    oo.set(p)
    assert p is i.get()


def test_configvar():
    t = MockTaskForVar()
    s = ConfigSpec("a", dict, True, True, None)
    c = _ConfigVar(t, s)
    raises(AssertionError, lambda: c.get())  # required not set

    p = ParamDict()
    s = ConfigSpec("a", dict, True, False, p)
    c = _ConfigVar(t, s)
    assert p is c.get()
    c.set(None)
    assert c.get() is None

    p = ParamDict()
    s = ConfigSpec("a", ParamDict, False, False, p)
    c = _ConfigVar(t, s)
    assert p is c.get()
    raises(AssertionError, lambda: c.set(None))
    assert p is c.get()

    p2 = ParamDict()
    s2 = ConfigSpec("x", dict, False, False, p2)
    c2 = _ConfigVar(t, s2)
    assert p2 is c2.get()  # not set, use the defaut
    c2.set_dependency(c)  # set parent
    assert p is c2.get()  # get parent value
    p3 = ParamDict()
    c.set(p3)  # set on parent will change child get
    assert p3 is c.get()
    assert p3 is c2.get()


def test_dependencydict():
    t = MockTaskForVar()
    s = ConfigSpec("a", int, True, False, 1)
    c1 = _ConfigVar(t, s)
    s = ConfigSpec("b", int, True, False, 2)
    c2 = _ConfigVar(t, s)
    d = _DependencyDict(IndexedOrderedDict([("a", c1), ("b", c2)]))
    assert 2 == len(d)
    assert 1 == d["a"]
    assert 2 == d["b"]
    c2.set(3)
    assert 3 == d["b"]
    assert [("a", 1), ("b", 3)] == list(d.items())
    with raises(InvalidOperationError):
        d["c"] = 1
    with raises(InvalidOperationError):
        d["b"] = 1
    with raises(InvalidOperationError):
        d.update(dict())
    assert 3 == d["b"]
    assert "3" == d.get_or_throw("b", str)
    assert "3" == d.get("b", "x")
    assert 0 == d.get("d", 0)
    with raises(KeyError):
        d.get_or_throw("d", str)


def _dummy(a: int, b: str) -> float:
    return 0.0


def test_cast():
    x = cast(Callable[[int], float], _dummy)
    print(x)


def _mock_task_func():
    pass


class MockSpec(object):
    @property
    def deterministic(self):
        return False


class MockTaskForVar(_Task):
    def __init__(self):
        self.spec = MockSpec()

    @property
    def name(self) -> str:
        return "taskname"

    def __uuid__(self) -> str:
        return "id"
